@@ -9,7 +9,7 @@ from requests_auth.errors import *
 logger = logging.getLogger(__name__)
 
 
-def decode_base64(base64_encoded_string: str) -> str:
+def _decode_base64(base64_encoded_string: str) -> str:
     """
     Decode base64, padding being optional.
 
@@ -22,8 +22,18 @@ def decode_base64(base64_encoded_string: str) -> str:
     return base64.b64decode(base64_encoded_string).decode("unicode_escape")
 
 
-def is_expired(expiry: float) -> bool:
-    return datetime.datetime.utcfromtimestamp(expiry) < datetime.datetime.utcnow()
+def _is_expired(expiry: float, early_expiry: float) -> bool:
+    return (
+        datetime.datetime.utcfromtimestamp(expiry - early_expiry)
+        < datetime.datetime.utcnow()
+    )
+
+
+def _to_expiry(expires_in: Union[int, str]) -> float:
+    expiry = datetime.datetime.utcnow().replace(
+        tzinfo=datetime.timezone.utc
+    ) + datetime.timedelta(seconds=int(expires_in))
+    return expiry.timestamp()
 
 
 class TokenMemoryCache:
@@ -36,7 +46,7 @@ class TokenMemoryCache:
         self.forbid_concurrent_cache_access = threading.Lock()
         self.forbid_concurrent_missing_token_function_call = threading.Lock()
 
-    def add_bearer_token(self, key: str, token: str):
+    def _add_bearer_token(self, key: str, token: str):
         """
         Set the bearer token and save it
         :param key: key identifier of the token
@@ -48,14 +58,20 @@ class TokenMemoryCache:
             raise InvalidToken(token)
 
         header, body, other = token.split(".")
-        body = json.loads(decode_base64(body))
+        body = json.loads(_decode_base64(body))
         expiry = body.get("exp")
         if not expiry:
             raise TokenExpiryNotProvided(expiry)
 
         self._add_token(key, token, expiry)
 
-    def add_access_token(self, key: str, token: str, expires_in: int, refresh_token: str = None):
+    def _add_access_token(
+        self,
+        key: str,
+        token: str,
+        expires_in: Union[int, str],
+        refresh_token: str = None,
+    ):
         """
         Set the bearer token and save it
         :param key: key identifier of the token
@@ -64,12 +80,11 @@ class TokenMemoryCache:
         :param refresh_token: refresh token value
         :raise InvalidToken: In case token is invalid.
         """
-        expiry = datetime.datetime.utcnow().replace(
-            tzinfo=datetime.timezone.utc
-        ) + datetime.timedelta(seconds=int(expires_in))
-        self._add_token(key, token, expiry.timestamp(), refresh_token)
+        self._add_token(key, token, _to_expiry(expires_in), refresh_token)
 
-    def _add_token(self, key: str, token: str, expiry: float, refresh_token: str = None):
+    def _add_token(
+        self, key: str, token: str, expiry: float, refresh_token: str = None
+    ):
         """
         Set the bearer token and save it
         :param key: key identifier of the token
@@ -84,13 +99,26 @@ class TokenMemoryCache:
                 f'Inserting token expiring on {datetime.datetime.utcfromtimestamp(expiry)} (UTC) with "{key}" key: {token}'
             )
 
-    def get_token(self, key: str, on_missing_token=None, *on_missing_token_args, on_expired_token=None) -> str:
+    def get_token(
+        self,
+        key: str,
+        *,
+        early_expiry: float = 30.0,
+        on_missing_token=None,
+        on_expired_token=None,
+        **on_missing_token_kwargs,
+    ) -> str:
         """
         Return the bearer token.
         :param key: key identifier of the token
+        :param early_expiry: As the time between the token extraction from cache and the token reception on server side
+        might not higher than one second, on slow networks, token might be expired when received by the actual server,
+        even if still valid when fetched.
+        This is the number of seconds to subtract to the actual token expiry. Token will be considered as
+        expired 30 seconds before real expiry by default.
         :param on_missing_token: function to call when token is expired or missing (returning token and expiry tuple)
-        :param on_missing_token_args: arguments of the function
         :param on_expired_token: function to call to refresh the token when it is expired
+        :param on_missing_token_kwargs: arguments of the on_missing_token function (key-value arguments)
         :return: the token
         :raise AuthenticationFailed: in case token cannot be retrieved.
         """
@@ -104,7 +132,7 @@ class TokenMemoryCache:
                     bearer, expiry = token
                 else:
                     bearer, expiry, refresh_token = token
-                if is_expired(expiry):
+                if _is_expired(expiry, early_expiry):
                     logger.debug(f'Authentication token with "{key}" key is expired.')
                     del self.tokens[key]
                 else:
@@ -116,13 +144,17 @@ class TokenMemoryCache:
         if refresh_token is not None and on_expired_token is not None:
             try:
                 with self.forbid_concurrent_missing_token_function_call:
-                    state, token, expires_in, refresh_token = on_expired_token(refresh_token)
-                    self.add_access_token(state, token, expires_in, refresh_token)
+                    state, token, expires_in, refresh_token = on_expired_token(
+                        refresh_token
+                    )
+                    self._add_access_token(state, token, expires_in, refresh_token)
                     logger.debug(f"Refreshed token with key {key}.")
                 with self.forbid_concurrent_cache_access:
                     if state in self.tokens:
                         bearer, expiry, refresh_token = self.tokens[state]
-                        logger.debug(f"Using newly refreshed token, expiring on {datetime.datetime.utcfromtimestamp(expiry)} (UTC).")
+                        logger.debug(
+                            f"Using newly refreshed token, expiring on {datetime.datetime.utcfromtimestamp(expiry)} (UTC)."
+                        )
                         return bearer
             except (InvalidGrantRequest, GrantNotProvided):
                 logger.debug(f"Failed to refresh token.")
@@ -130,16 +162,16 @@ class TokenMemoryCache:
         logger.debug("Token cannot be found in cache.")
         if on_missing_token is not None:
             with self.forbid_concurrent_missing_token_function_call:
-                new_token = on_missing_token(*on_missing_token_args)
+                new_token = on_missing_token(**on_missing_token_kwargs)
                 if len(new_token) == 2:  # Bearer token
                     state, token = new_token
-                    self.add_bearer_token(state, token)
+                    self._add_bearer_token(state, token)
                 elif len(new_token) == 3:  # Access token
                     state, token, expires_in = new_token
-                    self.add_access_token(state, token, expires_in)
+                    self._add_access_token(state, token, expires_in)
                 else:  # Access token and Refresh token
                     state, token, expires_in, refresh_token = new_token
-                    self.add_access_token(state, token, expires_in, refresh_token)
+                    self._add_access_token(state, token, expires_in, refresh_token)
                 if key != state:
                     logger.warning(
                         f"Using a token received on another key than expected. Expecting {key} but was {state}."
